---
title: Log Event
description: A detailed guide on Vector's internal log data model.
---

import SVG from 'react-inlinesvg';

<SVG src="/img/data-model-log.svg" />

<!--
     THIS FILE IS AUTOGENERATED!

     To make changes please edit the template located at:

     website/docs/about/data-model/log.md.erb
-->

## Description

A `log` event is a structured represention of a point-in-time event. It contains
an arbitrary set of fields (key/value pairs) that describe the event.

## Examples

import Tabs from '@theme/Tabs';
import TabItem from '@theme/TabItem';

<Tabs
  block={true}
  defaultValue="default"
  values={[
    { label: 'Default Schema', value: 'default', },
    { label: 'Custom Fields', value: 'custom', },
    { label: 'Nested Fields', value: 'nested', },
  ]
}>
<TabItem value="default">

```js
{
  "host": "my.host.com",
  "message": "<13>Feb 13 20:07:26 74794bfb6795 root[8539]: i am foobar",
  "timestamp": "2019-11-01T21:15:47+00:00"
}
```

</TabItem>
<TabItem value="custom">

```js
{
  "host": "my.host.com",
  "message": "<13>Feb 13 20:07:26 74794bfb6795 root[8539]: i am foobar",
  "timestamp": "2019-11-01T21:15:47+00:00",
  "custom_field": "custom_value" // custom keys are allowed
}
```

</TabItem>
<TabItem value="nested">

```js
{
  "host": "my.host.com",
  "message": "<13>Feb 13 20:07:26 74794bfb6795 root[8539]: i am foobar",
  "timestamp": "2019-11-01T21:15:47+00:00",
  "parent.child": "child_value" // nested fields are delimited with a `.` char
}
```

</TabItem>
</Tabs>

## Schema

import Alert from '@site/src/components/Alert';

<Alert type="info">

The following fields are the _default_ schema. All of these field names can be
changed via the
[global `log_schema` options][docs.reference.global-options#log_schema].

</Alert>

import Fields from '@site/src/components/Fields';

import Field from '@site/src/components/Field';

<Fields filters={true}>


<Field
  common={false}
  defaultValue={null}
  enumValues={null}
  examples={["my.host.com"]}
  groups={[]}
  name={"host"}
  path={null}
  relevantWhen={null}
  required={false}
  templateable={false}
  type={"string"}
  unit={null}
  >

### host

Represents the originating host of the log. This is automatically set within select [sources][docs.sources] if the key does not exist. Change this field name via the [global `host_key` option][docs.reference.global-options#host_key] or the source-level `host_key` option for relevant sources. See [Changing The Default Schema](#changing-the-default-schema) for more info.


</Field>


<Field
  common={true}
  defaultValue={null}
  enumValues={null}
  examples={["<13>Feb 13 20:07:26 74794bfb6795 root[8539]: i am foobar"]}
  groups={[]}
  name={"message"}
  path={null}
  relevantWhen={null}
  required={true}
  templateable={false}
  type={"string"}
  unit={null}
  >

### message

Represents the log message. Change this field name via the [global `message_key` option][docs.reference.global-options#message-key] or the source-level `message_key` option for relevant sources.
 See [Changing The Default Schema](#changing-the-default-schema) for more info.


</Field>


<Field
  common={true}
  defaultValue={null}
  enumValues={null}
  examples={["2019-11-01T21:15:47+00:00"]}
  groups={[]}
  name={"timestamp"}
  path={null}
  relevantWhen={null}
  required={true}
  templateable={false}
  type={"timestamp"}
  unit={null}
  >

### timestamp

A normalized [Rust DateTime struct][urls.rust_date_time] in UTC. Change this field name via the [global `timestamp_key` option][docs.reference.global-options#message-key] or the source-level `timestamp_key` option for relevant sources.
 See [Changing The Default Schema](#changing-the-default-schema) and [Timestamp Coercion](#timestamp-coercion) for more info.


</Field>


<Field
  common={false}
  defaultValue={null}
  enumValues={null}
  examples={[{"my-key":"my-value"},{"parent":{"child":"child-value"}}]}
  groups={[]}
  name={"`[custom-key]`"}
  path={null}
  relevantWhen={null}
  required={false}
  templateable={false}
  type={"*"}
  unit={null}
  >

### `[custom-key]`

In addition to the defined fields, a log event can have any number of additional fields. This includes nested fields.



</Field>


</Fields>

## Components

import Jump from '@site/src/components/Jump';

<Jump to="/components/?log=true">View all log compatible components</Jump>

## How It Works

### Changing The Default Schema

Vector is unique in that you are not required to use a specific schema. You
can change the default schema. This makes integrating Vector into existing
pipelines much easier. The [`message`](#message), [`host`](#host), and [`timestamp`](#timestamp) field names can
all be changed via the
[global `log_schema` options][docs.reference.global-options#log_schema].

#### Field name collisions

When you send structured data to Vector your field names always take precedence.
For example, if you send a JSON object with a [`timestamp`](#timestamp) key, Vector will
_not_ override the value of that field. Vector will only set that field if it
is not present.

### Time Zones

If Vector receives a timestamp that does not contain timezone information
Vector assumes the timestamp is in local time, and will convert the timestamp
to UTC from the local time.

<Alert type="warning">

It is important that the host system contain time zone data files to properly
determine the local time zone. This is typically installed through the `tzdata`
package. See [issue 551][urls.issue_551] for more info.

</Alert>

### Timestamp Coercion

There are cases where Vector interacts with formats that do not have a formal
timestamp defintion, such as JSON. In these cases, Vector will ingest the
timestamp in it's primitive form (string or integer). You can then coerce the
field into a [`timestamp`](#timestamp) using the
[`coercer` transform][docs.transforms.coercer]. If you are parsing this data
out of a string, all Vector parser transforms include a `types` option,
allowing you to extract and coerce in one step.

### Types

#### Strings

Strings are UTF8 compatible and are only bounded by the available system
memory.

#### Ints

Integers are signed integers up to 64 bits.

#### Floats

Floats are signed floats up to 64 bits.

#### Booleans

Booleans represent binary true/false values.

#### Timestamps

Timestamps are represented as [`DateTime` Rust structs][urls.rust_date_time]
stored as UTC.

#### Null Values

 For compatibility with JSON log events, Vector also supports `null` values.


[docs.reference.global-options#host_key]: /docs/reference/global-options/#host_key
[docs.reference.global-options#log_schema]: /docs/reference/global-options/#log_schema
[docs.reference.global-options#message-key]: /docs/reference/global-options/#message-key
[docs.sources]: /docs/reference/sources/
[docs.transforms.coercer]: /docs/reference/transforms/coercer/
[urls.issue_551]: https://github.com/timberio/vector/issues/551
[urls.rust_date_time]: https://docs.rs/chrono/0.4.0/chrono/struct.DateTime.html
